/*
 *$Id: sct_ctrl_lib.c,v 1.9 2001/09/14 15:01:45 chooga Exp $
 */
#include "sct_ctrl_lib.h"

/* get the module's sysctl defines */
#include "sct_rules_module/module_interface.h"

#include <malloc.h>
#include <assert.h>
#include <stdio.h>
#include <string.h>

#include <linux/unistd.h>
#include <linux/sysctl.h>

struct sct_ctrl_type {
    /* empty, for now */
    int dummy;
};

_syscall1(int, _sysctl, struct __sysctl_args *, args);

/* init the library */
sct_ctrl_t
sct_ctrl_init(void)
{
    sct_ctrl_t ctrl = malloc(sizeof(*ctrl));
    if (!ctrl){
	return NULL;
    }

    memset(ctrl, 0, sizeof(*ctrl));
    return ctrl;
}

/* shutdown the library */
int sct_ctrl_cleanup(sct_ctrl_t ctrl)
{
    if (ctrl)
	free(ctrl);

    return 1;
}

/* returns 0 on error and 1 on success */
static int do_sysctl(sct_ctrl_t ctrl, unsigned char* buf, 
		     unsigned int buflen, int* sysctl_cmd, 
		     unsigned int sysctl_cmd_len)
{
    struct __sysctl_args args =
	{ sysctl_cmd, sysctl_cmd_len, NULL, 0, buf, buflen};

    int rc;

    assert (sysctl_cmd);

    if (!sysctl_cmd)
	return 0;

    rc = _sysctl(&args);
    if (rc != 0)
        fprintf(stderr, "sysctl returned '%d'\n", rc);

    return (rc != -1) ? 1 : 0;
}

static int
sct_ctrl_common(sct_ctrl_t ctrl, unsigned int sysctl_command, 
		int rule_id, int syscall_id, 
		filter_node_t* filter, action_t* act,
		int enable_rule)
{
    /* we ignore enable_rule for now */

    int serial_len;
    unsigned char buf[1000];
    int sysctl_cmd_struct[] = { CTL_TRACKER, sysctl_command };

    assert(rule_id >= 0 && syscall_id >= 0);

    if (!(rule_id >= 0) || !(syscall_id >= 0))
	return 0;
    
    serial_len = 
	serialize_tracking_rule(rule_id, syscall_id, act, 
				filter, buf, sizeof(buf));

    if (serial_len <= 0){
	fprintf(stderr, "failed to serialize tracking rule, "
		"error '%d'\n", serial_len);
	return 0;
    }

    return do_sysctl(ctrl, buf, serial_len, sysctl_cmd_struct, 
		     sizeof(sysctl_cmd_struct)/ sizeof(sysctl_cmd_struct[0]));
}

		     

/* add rule */
int sct_ctrl_add_rule(sct_ctrl_t ctrl, int rule_id, int syscall_id, 
		      filter_node_t* filter, action_t* act,
		      int enable_rule)
{
    assert (filter && act); 
    if (!filter || !act)
	return 0;

    return sct_ctrl_common(ctrl, CTL_TRACKER_ADD_RULE, 
			   rule_id, syscall_id, 
			   filter, act, enable_rule);
}

/* delete a rule */
int
sct_ctrl_del_rule(sct_ctrl_t ctrl, int rule_id, int syscall_id)
{
    return sct_ctrl_common(ctrl, CTL_TRACKER_DEL_RULE, 
			   rule_id, syscall_id, 
			   NULL, NULL, 0);
}

/* enable a rule */
int 
sct_ctrl_enable_rule(sct_ctrl_t ctrl, int rule_id, int syscall_id)
{
    fprintf(stderr, "eek! sct_ctrl_enable_rule is unimplented!\n");
    return 0;
}

/* disable a rule */
int 
sct_ctrl_disable_rule(sct_ctrl_t ctrl, int rule_id, int syscall_id)
{
    fprintf(stderr, "eek! sct_ctrl_disable_rule is unimplented!\n");
    return 0;
}

/* delete all rules for this syscall */
int 
sct_ctrl_del_all_syscall_rules(sct_ctrl_t ctrl, int syscall_id)
{
    return sct_ctrl_common(ctrl, CTL_TRACKER_DEL_ALL_RULES_FOR_SYSCALL, 
			   /* rule_id */ 0, syscall_id, 
			   NULL, NULL, 0);
}

/* delete all rules for ALL syscalls */
int 
sct_ctrl_del_all_rules(sct_ctrl_t ctrl)
{
    return sct_ctrl_common(ctrl, CTL_TRACKER_DEL_ALL_MODULE_RULES, 
			   /* rule_id */ 0, /* syscall_id */ 0,
			   NULL, NULL, 0);
}

/* print the current rules */
int
sct_ctrl_print_all_rules(sct_ctrl_t ctrl)
{
    return sct_ctrl_common(ctrl, CTL_TRACKER_PRINT_ALL_RULES, 
			   /* rule_id */ 0, /* syscall_id */ 0,
			   NULL, NULL, 0);
}




