/*
 *$Id: encode_decode.c,v 1.1 2001/09/14 14:58:43 chooga Exp $
 */
/*
 * Rule engine - primitive types encoding/decoding functions.
 */

#include "mem_manager_public.h"
#include "archdep.h"

/* encode the given unsigned char at the given pos, */
/* as long as the buffer is not too small.          */
int
encode_char(unsigned char* buf,
	    int buflen,
	    unsigned char** pos,
	    unsigned char c)
{
	int rc = 0;

	if (*pos - buf <= buflen - sizeof(char)) {
		(**pos) = c;
		(*pos)++;
		rc = 1;
	}
	else {
		rc = -1;
	}

	return rc;
}

/* decode the given unsigned char at the given pos, */
/* as long as we are not past end of buf.           */
int
decode_char(unsigned char* buf,
	    int buflen,
	    unsigned char** pos,
	    unsigned char* pc)
{
	int rc = 0;

	if (*pos - buf <= buflen - sizeof(char)) {
		*pc = (**pos);
		(*pos)++;
		rc = 1;
	}
	else {
		rc = -1;
	}

	return rc;
}

/* encode the given int at the given pos,  */
/* as long as the buffer is not too small. */
int
encode_int(unsigned char* buf,
	   int buflen,
	   unsigned char** pos,
	   int i)
{
	int rc = 0;

	if (*pos - buf <= buflen - sizeof(int)) {
		memcpy(*pos, &i, sizeof(int));
		(*pos) += sizeof(int);
		rc = 1;
	}
	else {
		rc = -1;
	}

	return rc;
}

/* decode the given int at the given pos, */
/* as long as we are not past end of buf. */
int
decode_int(unsigned char* buf,
	   int buflen,
	   unsigned char** pos,
	   int* pi)
{
	int rc = 0;

	if (*pos - buf <= buflen - sizeof(int)) {
		memcpy(pi, *pos, sizeof(int));
		(*pos) += sizeof(int);
		rc = 1;
	}
	else {
		rc = -1;
	}

	return rc;
}

/* encode the given long at the given pos, */
/* as long as the buffer is not too small. */
int
encode_long(unsigned char* buf,
	    int buflen,
	    unsigned char** pos,
	    long l)
{
	int rc = 0;

	if (*pos - buf <= buflen - sizeof(long)) {
		memcpy(*pos, &l, sizeof(long));
		(*pos) += sizeof(long);
		rc = 1;
	}
	else {
		rc = -1;
	}

	return rc;
}

/* decode the given long at the given pos, */
/* as long aswe are not past end of buf.   */
int
decode_long(unsigned char* buf,
	    int buflen,
	    unsigned char** pos,
	    long* pl)
{
	int rc = 0;

	if (*pos - buf <= buflen - sizeof(long)) {
		memcpy(pl, *pos, sizeof(long));
		(*pos) += sizeof(long);
		rc = 1;
	}
	else {
		rc = -1;
	}

	return rc;
}

/* encode the given short at the given pos, */
/* as long as the buffer is not too small.  */
int
encode_short(unsigned char* buf,
	     int buflen,
	     unsigned char** pos,
	     short s)
{
	int rc = 0;

	if (*pos - buf <= buflen - sizeof(short)) {
		memcpy(*pos, &s, sizeof(short));
		(*pos) += sizeof(short);
		rc = 1;
	}
	else {
		rc = -1;
	}

	return rc;
}

/* decode the given short at the given pos, */
/* as long as we are not past end of buf.   */
int
decode_short(unsigned char* buf,
	     int buflen,
	     unsigned char** pos,
	     short* ps)
{
	int rc = 0;

	if (*pos - buf <= buflen - sizeof(short)) {
		memcpy(ps, *pos, sizeof(short));
		(*pos) += sizeof(short);
		rc = 1;
	}
	else {
		rc = -1;
	}

	return rc;
}

/* encode the given float at the given pos, */
/* as long as the buffer is not too small.  */
int
encode_float(unsigned char* buf,
	     int buflen,
	     unsigned char** pos,
	     float f)
{
	int rc = 0;

	if (*pos - buf <= buflen - sizeof(float)) {
		memcpy(*pos, &f, sizeof(float));
		(*pos) += sizeof(float);
		rc = 1;
	}
	else {
		rc = -1;
	}

	return rc;
}

/* decode the given float at the given pos, */
/* as long as we are nto past end of buf.   */
int
decode_float(unsigned char* buf,
	     int buflen,
	     unsigned char** pos,
	     float* pf)
{
	int rc = 0;

	if (*pos - buf <= buflen - sizeof(float)) {
		memcpy(pf, *pos, sizeof(float));
		(*pos) += sizeof(float);
		rc = 1;
	}
	else {
		rc = -1;
	}

	return rc;
}

/* encode the given string at the given pos, */
/* as long as the buffer is not too small.   */
int
encode_string(unsigned char* buf,
	      int buflen,
	      unsigned char** pos,
	      const char* str)
{
	int rc = 0;
	int len = strlen(str);

	/* encode the string's len, and then the string - */
	/* withut the string's delimiting '\0' character. */
	if (*pos - buf <= buflen - (sizeof(int) + len)) {
		memcpy(*pos, &len, sizeof(int));
		(*pos) += sizeof(int);
		memcpy(*pos, str, len);
		(*pos) += len;
		rc = 1;
	}
	else {
		rc = -1;
	}

	return rc;
}

/* decode the given string at the given pos, */
/* as long as we are not past end of buf.    */
int
decode_string(unsigned char* buf,
	      int buflen,
	      unsigned char** pos,
	      char** pstr)
{
	int rc = 0;
	int len;

	/* decode the string's len, and then the string -    */
	/* without the string's delimiting '\0' character.   */
	/* finally, append a '\0' char to the end of string. */
	rc = decode_int(buf, buflen, pos, &len);
	if (rc > 0) {
		if (*pos - buf <= buflen - len) {
			(*pstr) = mem_alloc(len+1);
			if (*pstr) {
				memcpy(*pstr, *pos, len);
				(*pos) += len;
				*(*pstr + len) = '\0';
				rc = 1;
			}
			else {
				rc = 0;
			}
		}
		else {
			rc = -1;
		}
	}

	return rc;
}

