/*
 *$Id: filter_eval.c,v 1.1 2001/09/14 14:58:43 chooga Exp $
 */
/*
 * Rule engine - implementation of filter node evaluation.
 */

#include "mem_manager_public.h"
#include "rule_engine_private.h"
#include "filter_vars_public.h"

#include "archdep.h"
/*
 * forward declarations.
 */
extern int do_eval_filter_node(struct tracker_filter_node* f,
		    	       struct tracker_filter_node* result);

/*
 * Private macros.
 */

/* typed result macros. */
#define	RESULT_DATA_TYPE(res)	((res)->u.constant.type)
#define	RESULT_BOOL(res)	((res)->u.constant.u.b)
#define	RESULT_CHAR(res)	((res)->u.constant.u.c)
#define	RESULT_SHORT(res)	((res)->u.constant.u.s)
#define	RESULT_INT(res)		((res)->u.constant.u.i)
#define	RESULT_LONG(res)	((res)->u.constant.u.l)
#define	RESULT_FLOAT(res)	((res)->u.constant.u.f)
#define	RESULT_STRING(res)	((res)->u.constant.u.str)


/*
 * Private functions.
 */

/* number of bytes each data type occupies.   */
/* must be in sycn with the data_type_t enum. */
static int data_type_sizes[] = {
	1,		/* DT_BOOL   */
	sizeof(char),	/* DT_CHAR   */
	sizeof(int),	/* DT_INT    */
	sizeof(short),	/* DT_SHORT  */
	sizeof(long),	/* DT_LONG   */
	sizeof(float),	/* DT_FLOAT  */
	sizeof(char*)	/* DT_STRING */
};

/* compare two data types. return 0 if they are equal, -1  */
/* if the first takes more space, 1 if the second takes    */
/* more space.                                             */
int
typecmp(data_type_t t1, data_type_t t2)
{
	int ts1 = 0;
	int ts2 = 0;

	if (t1 >= 0 && t1 < DT_LAST_TYPE) {
		ts1 = data_type_sizes[t1];
	}
	if (t2 >= 0 && t2 < DT_LAST_TYPE) {
		ts2 = data_type_sizes[t2];
	}

	return (ts1 < ts2 ? -1 : (ts1 == ts2 ? 0 : 1));
}

/* cast a constant value into a boolean value, similarly   */
/* to how this is doen in C. Exception for a string value. */
int
cast_bool(struct tracker_filter_node* val)
{
	int result = TF_FALSE;

	switch(CONST_DATA_TYPE(val)) {
		case DT_BOOL:
			result = CONST_BOOL(val);
			break;
		case DT_CHAR:
			result = CONST_CHAR(val) ? TF_TRUE : TF_FALSE;
			break;
		case DT_SHORT:
			result = CONST_SHORT(val) ? TF_TRUE : TF_FALSE;
			break;
		case DT_INT:
			result = CONST_INT(val) ? TF_TRUE : TF_FALSE;
			break;
		case DT_LONG:
			result = CONST_LONG(val) ? TF_TRUE : TF_FALSE;
			break;
		case DT_FLOAT:
			result = CONST_FLOAT(val) ? TF_TRUE : TF_FALSE;
			break;
		case DT_STRING:
			result = CONST_STRING(val) ? TF_TRUE : TF_FALSE;
			break;
		default:
			/* shouldn't happen. */
			result = 0;
			break;
	}

	return result;
}

/* cast a constant value into a char value, similarly      */
/* to how this is doen in C. Exception for a string value. */
char
cast_char(struct tracker_filter_node* val)
{
	char result = 0;

	switch(CONST_DATA_TYPE(val)) {
		case DT_BOOL:
			result = (char)CONST_BOOL(val);
			break;
		case DT_CHAR:
			result = CONST_CHAR(val);
			break;
		case DT_SHORT:
			result = (char)CONST_SHORT(val);
			break;
		case DT_INT:
			result = (char)CONST_INT(val);
			break;
		case DT_LONG:
			result = (char)CONST_LONG(val);
			break;
		case DT_FLOAT:
			result = (char)CONST_FLOAT(val);
			break;
		case DT_STRING:
			/* TODO - how is a string cast to a char? */
			result = 0;
			break;
		default:
			/* shouldn't happen. */
			result = '\0';
			break;
	}

	return result;
}

/* cast a constant value into a short value, similarly     */
/* to how this is doen in C. Exception for a string value. */
short
cast_short(struct tracker_filter_node* val)
{
	short result = 0;

	switch(CONST_DATA_TYPE(val)) {
		case DT_BOOL:
			result = CONST_BOOL(val);
			break;
		case DT_CHAR:
			result = CONST_CHAR(val);
			break;
		case DT_SHORT:
			result = CONST_SHORT(val);
			break;
		case DT_INT:
			result = (short)CONST_INT(val);
			break;
		case DT_LONG:
			result = (short)CONST_LONG(val);
			break;
		case DT_FLOAT:
			result = (short)CONST_FLOAT(val);
			break;
		case DT_STRING:
			/* TODO - write an atoi function (for kernel mode) */
			/*        and use it here.                         */
			result = 0;
			break;
		default:
			/* shouldn't happen. */
			result = 0;
			break;
	}

	return result;
}
/* cast a constant value into an int value, similarly      */
/* to how this is doen in C. Exception for a string value. */
int
cast_int(struct tracker_filter_node* val)
{
	int result = 0;

	switch(CONST_DATA_TYPE(val)) {
		case DT_BOOL:
			result = CONST_BOOL(val);
			break;
		case DT_CHAR:
			result = CONST_CHAR(val);
			break;
		case DT_SHORT:
			result = CONST_SHORT(val);
			break;
		case DT_INT:
			result = CONST_INT(val);
			break;
		case DT_LONG:
			result = CONST_LONG(val);
			break;
		case DT_FLOAT:
			result = (int)CONST_FLOAT(val);
			break;
		case DT_STRING:
			/* TODO - write an atoi function (for kernel mode) */
			/*        and use it here.                         */
			result = 0;
			break;
		default:
			/* shouldn't happen. */
			result = 0;
			break;
	}

	return result;
}

/* cast a constant value into a long value, similarly      */
/* to how this is doen in C. Exception for a string value. */
long
cast_long(struct tracker_filter_node* val)
{
	long result = 0;

	switch(CONST_DATA_TYPE(val)) {
		case DT_BOOL:
			result = CONST_BOOL(val);
			break;
		case DT_CHAR:
			result = CONST_CHAR(val);
			break;
		case DT_SHORT:
			result = CONST_SHORT(val);
			break;
		case DT_INT:
			result = CONST_INT(val);
			break;
		case DT_LONG:
			result = CONST_LONG(val);
			break;
		case DT_FLOAT:
			result = (long)CONST_FLOAT(val);
			break;
		case DT_STRING:
			/* TODO - write an atoi function (for kernel mode) */
			/*        and use it here.                         */
			result = 0;
			break;
		default:
			/* shouldn't happen. */
			result = 0;
			break;
	}

	return result;
}

/* cast a constant value into a string value, similarly    */
/* to how this is doen in C. Strings are strings. numbers  */
/* are being stringified.                                  */
/* the caller should free the resulting string.            */
char*
cast_string(struct tracker_filter_node* val)
{
	char* result = 0;

	switch(CONST_DATA_TYPE(val)) {
		case DT_BOOL:
			result = mem_alloc(1+1);
			result[0] = (CONST_BOOL(val) == 1 ? '1' : '0');
			result[1] = '\0';
			break;
		case DT_CHAR:
			result = mem_alloc(1+1);
			result[0] = CONST_CHAR(val);
			result[1] = '\0';
			break;
		case DT_SHORT:
			result = mem_alloc(5+1);
			sprintf(result, "%d", CONST_SHORT(val));
			break;
		case DT_INT:
			result = mem_alloc(20+1);
			sprintf(result, "%d", CONST_INT(val));
			break;
		case DT_LONG:
			result = mem_alloc(20+1);
			sprintf(result, "%ld", CONST_LONG(val));
			break;
		case DT_FLOAT:
			result = mem_alloc(50+1);
			sprintf(result, "%f", CONST_FLOAT(val));
			break;
		case DT_STRING:
			result = mem_alloc(strlen(CONST_STRING(val))+1);
			strcpy(result, CONST_STRING(val));
			break;
		default:
			/* shouldn't happen. */
			result = 0;
			break;
	}

	return result;
}

/* check if the given string matches the givven pattern. */
/* currently implements a very trivial strstr matching.  */
/* TODO - make real pattern matching.                    */
int
pattern_match(const char* str, const char* pat)
{
	if (str && pat && strstr(str, pat))
		return 1;
	else
		return 0;
}

/* evaluate the given constant node.		  */
exp_error_t
eval_filter_node_const(struct tracker_filter_node* f,
		       struct tracker_filter_node* result)
{
	exp_error_t rc = ERR_NONE;

	result->type = NT_CONST;
	result->data_type = f->data_type;
	RESULT_DATA_TYPE(result) = f->data_type;
	switch (result->data_type) {
		case DT_BOOL:
			RESULT_BOOL(result) = CONST_BOOL(f);
			break;
        	case DT_CHAR:
			RESULT_CHAR(result) = CONST_CHAR(f);
			break;
        	case DT_INT:
			RESULT_INT(result) = CONST_INT(f);
			break;
        	case DT_SHORT:
			RESULT_SHORT(result) = CONST_SHORT(f);
			break;
        	case DT_LONG:
			RESULT_LONG(result) = CONST_LONG(f);
			break;
        	case DT_FLOAT:
			RESULT_FLOAT(result) = CONST_FLOAT(f);
			break;
        	case DT_STRING:
			{
				const char* str = CONST_STRING(f);
				if (str) {
					RESULT_STRING(result) =
						mem_alloc(strlen(str)+1);
					strcpy(RESULT_STRING(result), str);
				}
				else {
					RESULT_STRING(result) = NULL;
				}
			}
			break;
		default:
			/* shouldn't happen. */
			rc = ERR_INVALID_DATA_TYPE;
			break;
	}

	return rc;
}

/* evaluate the given variable node.		  */
exp_error_t
eval_filter_node_var(struct tracker_filter_node* f,
		     struct tracker_filter_node* result)
{
	exp_error_t rc = ERR_NONE;
	filter_node_t* var_val = NULL;

	result->type = NT_CONST;
	result->data_type = f->data_type;
	RESULT_DATA_TYPE(result) = f->data_type;

	switch (f->u.var.type) {
		case VT_PARAM:
			var_val = get_var_indexed(f->u.var.type,
						  f->u.var.var_index);
			break;
		default:
			var_val = get_var(f->u.var.type);
			break;
	}

	if (var_val) {
		rc = eval_filter_node_const(var_val, result);
	}
	else {
		rc = ERR_INVALID_VAR;
	}

	return rc;
}


/*
 * Evaluating operators. All 'values' are constants.
 */

exp_error_t
eval_op_equal(struct tracker_filter_node* f_op,
	      struct tracker_filter_node* val1,
	      struct tracker_filter_node* val2,
	      struct tracker_filter_node* result)
{
	exp_error_t rc = ERR_NONE;
	data_type_t t1 = val1->data_type;
	data_type_t t2 = val2->data_type;

	result->type = NT_CONST;
	RESULT_DATA_TYPE(result) = DT_BOOL;
	
	if (t1 == DT_STRING && t2 == DT_STRING) {
		const char* str1 = val1->u.constant.u.str;
		const char* str2 = val2->u.constant.u.str;
		if (!str1) {
			if (!str2) {
				RESULT_BOOL(result) = TF_TRUE;
			}
			else {
				RESULT_BOOL(result) = TF_FALSE;
			}
		}
		else {
			if (!str2) {
				RESULT_BOOL(result) = TF_FALSE;
			}
			else {
				RESULT_BOOL(result) =
					(!strcmp(str1, str2) ? TF_TRUE
							     : TF_FALSE);
			}
		}
	}
	else if (t1 == DT_STRING || t2 == DT_STRING) {
		/* one is a string and the other is not - type mismatch. */
		rc = ERR_TYPE_MISMATCH;
	}
	else {
		long l1 = cast_long(val1);
		long l2 = cast_long(val2);
		RESULT_BOOL(result) = (l1 == l2 ? TF_TRUE : TF_FALSE);
	}

	return rc;
}

exp_error_t
eval_op_less(struct tracker_filter_node* f_op,
	     struct tracker_filter_node* val1,
	     struct tracker_filter_node* val2,
	     struct tracker_filter_node* result)
{
	exp_error_t rc = ERR_NONE;
	data_type_t t1 = val1->data_type;
	data_type_t t2 = val2->data_type;

	result->type = NT_CONST;
	RESULT_DATA_TYPE(result) = DT_BOOL;
	
	if (t1 == DT_STRING && t2 == DT_STRING) {
		const char* str1 = val1->u.constant.u.str;
		const char* str2 = val2->u.constant.u.str;
		if (!str1) {
			if (!str2) {
				RESULT_BOOL(result) = TF_FALSE;
			}
			else {
				RESULT_BOOL(result) = TF_TRUE;
			}
		}
		else {
			if (!str2) {
				RESULT_BOOL(result) = TF_FALSE;
			}
			else {
				RESULT_BOOL(result) =
					(strcmp(str1, str2) < 0 ? TF_TRUE
							        : TF_FALSE);
			}
		}
	}
	else if (t1 == DT_STRING || t2 == DT_STRING) {
		/* one is a string and the other is not - type mismatch. */
		rc = ERR_TYPE_MISMATCH;
	}
	else {
		long l1 = cast_long(val1);
		long l2 = cast_long(val2);
		RESULT_BOOL(result) = (l1 < l2 ? TF_TRUE : TF_FALSE);
	}

	return rc;
}

exp_error_t
eval_op_greater(struct tracker_filter_node* f_op,
		struct tracker_filter_node* val1,
		struct tracker_filter_node* val2,
		struct tracker_filter_node* result)
{
	exp_error_t rc = ERR_NONE;
	data_type_t t1 = val1->data_type;
	data_type_t t2 = val2->data_type;

	result->type = NT_CONST;
	RESULT_DATA_TYPE(result) = DT_BOOL;
	
	if (t1 == DT_STRING && t2 == DT_STRING) {
		const char* str1 = val1->u.constant.u.str;
		const char* str2 = val2->u.constant.u.str;
		if (!str1) {
			if (!str2) {
				RESULT_BOOL(result) = TF_FALSE;
			}
			else {
				RESULT_BOOL(result) = TF_FALSE;
			}
		}
		else {
			if (!str2) {
				RESULT_BOOL(result) = TF_TRUE;
			}
			else {
				RESULT_BOOL(result) =
					(strcmp(str1, str2) > 0 ? TF_TRUE
							        : TF_FALSE);
			}
		}
	}
	else if (t1 == DT_STRING || t2 == DT_STRING) {
		/* one is a string and the other is not - type mismatch. */
		rc = ERR_TYPE_MISMATCH;
	}
	else {
		long l1 = cast_long(val1);
		long l2 = cast_long(val2);
		RESULT_BOOL(result) = (l1 > l2 ? TF_TRUE : TF_FALSE);
	}

	return rc;
}

exp_error_t
eval_op_less_equal(struct tracker_filter_node* f_op,
		   struct tracker_filter_node* val1,
		   struct tracker_filter_node* val2,
		   struct tracker_filter_node* result)
{
	/* less_equal is the same as !(greater) */
	exp_error_t rc = eval_op_greater(f_op, val1, val2, result);

	if (rc == ERR_NONE) {
		int eq_res = RESULT_BOOL(result);
		RESULT_BOOL(result) = (eq_res ? TF_FALSE : TF_TRUE);
	}

	return rc;
}

exp_error_t
eval_op_greater_equal(struct tracker_filter_node* f_op,
		      struct tracker_filter_node* val1,
		      struct tracker_filter_node* val2,
		      struct tracker_filter_node* result)
{
	/* greater_equal is the same as !(less) */
	exp_error_t rc = eval_op_less(f_op, val1, val2, result);

	if (rc == ERR_NONE) {
		int eq_res = RESULT_BOOL(result);
		RESULT_BOOL(result) = (eq_res ? TF_FALSE : TF_TRUE);
	}

	return rc;
}

exp_error_t
eval_op_not_equal(struct tracker_filter_node* f_op,
		  struct tracker_filter_node* val1,
		  struct tracker_filter_node* val2,
		  struct tracker_filter_node* result)
{
	/* not_equal is the same as !(equal) */
	exp_error_t rc = eval_op_equal(f_op, val1, val2, result);

	if (rc == ERR_NONE) {
		int eq_res = RESULT_BOOL(result);
		RESULT_BOOL(result) = (eq_res ? TF_FALSE : TF_TRUE);
	}

	return rc;
}

exp_error_t
eval_op_plus(struct tracker_filter_node* f_op,
	     struct tracker_filter_node* val1,
	     struct tracker_filter_node* val2,
	     struct tracker_filter_node* result)
{
	exp_error_t rc = ERR_NONE;
	data_type_t t1 = val1->data_type;
	data_type_t t2 = val2->data_type;

	result->type = NT_CONST;
	
	/* TODO - implement string concatenation here? */
	if (t1 == DT_STRING || t2 == DT_STRING) {
		rc = ERR_TYPE_MISMATCH;
	}
	else if (t1 == DT_BOOL || t2 == DT_BOOL) {
		/* cannot use '+' over bool input. */
		rc = ERR_TYPE_MISMATCH;
	}
	else {
		/* make sure the larger data type is in 'val1'. */
		int t2_is_larger = (typecmp(t1, t2) < 0);
		if (t2_is_larger) {
			/* swap the values and types. */
			data_type_t t_tmp = t1;
			struct tracker_filter_node* val_tmp = val1;
			val1 = val2;
			val2 = val_tmp;
			t1 = t2;
			t2 = t_tmp;
		}
		switch (t1) {
			case DT_LONG:
				RESULT_LONG(result) =
					CONST_LONG(val1) + cast_long(val2);
				result->data_type = DT_LONG;
				RESULT_DATA_TYPE(result) = DT_LONG;
				break;
			case DT_INT:
				RESULT_INT(result) =
					CONST_INT(val1) + cast_int(val2);
				result->data_type = DT_INT;
				RESULT_DATA_TYPE(result) = DT_INT;
				break;
			case DT_SHORT:
				RESULT_SHORT(result) =
					CONST_SHORT(val1) + cast_short(val2);
				result->data_type = DT_SHORT;
				RESULT_DATA_TYPE(result) = DT_SHORT;
				break;
			case DT_CHAR:
				RESULT_CHAR(result) =
					CONST_CHAR(val1) + cast_char(val2);
				result->data_type = DT_CHAR;
				RESULT_DATA_TYPE(result) = DT_CHAR;
				break;
			default:
				/* shouldn't happen. */
				rc = ERR_TYPE_MISMATCH;
				break;
		}
	}

	return rc;
}

exp_error_t
eval_op_minus(struct tracker_filter_node* f_op,
	      struct tracker_filter_node* val1,
	      struct tracker_filter_node* val2,
	      struct tracker_filter_node* result)
{
	exp_error_t rc = ERR_NONE;
	data_type_t t1 = val1->data_type;
	data_type_t t2 = val2->data_type;

	result->type = NT_CONST;
	
	/* TODO - implement string de-concatenation here? */
	if (t1 == DT_STRING || t2 == DT_STRING) {
		rc = ERR_TYPE_MISMATCH;
	}
	else if (t1 == DT_BOOL || t2 == DT_BOOL) {
		/* cannot use '-' over bool input. */
		rc = ERR_TYPE_MISMATCH;
	}
	else {
		/* make sure the larger data type is in 'val1'. */
		int t2_is_larger = (typecmp(t1, t2) < 0);
		if (t2_is_larger) {
			/* swap the values and types. */
			data_type_t t_tmp = t1;
			struct tracker_filter_node* val_tmp = val1;
			val1 = val2;
			val2 = val_tmp;
			t1 = t2;
			t2 = t_tmp;
		}
		switch (t1) {
			case DT_LONG:
				RESULT_LONG(result) =
					CONST_LONG(val1) - cast_long(val2);
				result->data_type = DT_LONG;
				RESULT_DATA_TYPE(result) = DT_LONG;
				break;
			case DT_INT:
				RESULT_INT(result) =
					CONST_INT(val1) - cast_int(val2);
				result->data_type = DT_INT;
				RESULT_DATA_TYPE(result) = DT_INT;
				break;
			case DT_SHORT:
				RESULT_SHORT(result) =
					CONST_SHORT(val1) - cast_short(val2);
				result->data_type = DT_SHORT;
				RESULT_DATA_TYPE(result) = DT_SHORT;
				break;
			case DT_CHAR:
				RESULT_CHAR(result) =
					CONST_CHAR(val1) - cast_char(val2);
				result->data_type = DT_CHAR;
				RESULT_DATA_TYPE(result) = DT_CHAR;
				break;
			default:
				/* shouldn't happen. */
				rc = ERR_TYPE_MISMATCH;
				break;
		}
	}

	return rc;
}

exp_error_t
eval_op_shift_left(struct tracker_filter_node* f_op,
		   struct tracker_filter_node* val1,
		   struct tracker_filter_node* val2,
		   struct tracker_filter_node* result)
{
	exp_error_t rc = ERR_NONE;
	data_type_t t1 = val1->data_type;
	data_type_t t2 = val2->data_type;

	/* we don't support shift operations for some data types. */
	if (t1 == DT_STRING || t1 == DT_BOOL || t2 == DT_STRING || t2 == DT_BOOL) {
		rc = ERR_TYPE_MISMATCH;
	}
	else {
		long ammount = cast_long(val2);
		result->type = NT_CONST;
		result->data_type = val1->data_type;
		RESULT_DATA_TYPE(result) = val1->data_type;

		switch (result->data_type) {
        		case DT_CHAR:
				RESULT_CHAR(result) <<= ammount;
				break;
        		case DT_INT:
				RESULT_INT(result) <<= ammount;
				break;
        		case DT_SHORT:
				RESULT_SHORT(result) <<= ammount;
				break;
        		case DT_LONG:
				RESULT_LONG(result) <<= ammount;
				break;
			default:
				/* shouldn't happen. */
				rc = ERR_INVALID_DATA_TYPE;
				break;
		}
	}

	return rc;
}

exp_error_t
eval_op_shift_right(struct tracker_filter_node* f_op,
		    struct tracker_filter_node* val1,
		    struct tracker_filter_node* val2,
		    struct tracker_filter_node* result)
{
	exp_error_t rc = ERR_NONE;
	data_type_t t1 = val1->data_type;
	data_type_t t2 = val2->data_type;

	/* we don't support shift operations for some data types. */
	if (t1 == DT_STRING || t1 == DT_BOOL || t2 == DT_STRING || t2 == DT_BOOL) {
		rc = ERR_TYPE_MISMATCH;
	}
	else {
		long ammount = cast_long(val2);
		result->type = NT_CONST;
		result->data_type = val1->data_type;
		RESULT_DATA_TYPE(result) = val1->data_type;

		switch (result->data_type) {
        		case DT_CHAR:
				RESULT_CHAR(result) >>= ammount;
				break;
        		case DT_INT:
				RESULT_INT(result) >>= ammount;
				break;
        		case DT_SHORT:
				RESULT_SHORT(result) >>= ammount;
				break;
        		case DT_LONG:
				RESULT_LONG(result) >>= ammount;
				break;
			default:
				/* shouldn't happen. */
				rc = ERR_INVALID_DATA_TYPE;
				break;
		}
	}

	return rc;
}

exp_error_t
eval_op_bit_and(struct tracker_filter_node* f_op,
		struct tracker_filter_node* val1,
		struct tracker_filter_node* val2,
		struct tracker_filter_node* result)
{
	exp_error_t rc = ERR_NONE;
	data_type_t t1 = val1->data_type;
	data_type_t t2 = val2->data_type;

	/* we don't support bit operations for some data types. */
	if (t1 == DT_STRING || t1 == DT_BOOL || t1 == DT_FLOAT) {
		rc = ERR_TYPE_MISMATCH;
	}
	if (t2 == DT_STRING || t2 == DT_BOOL || t2 == DT_FLOAT) {
		rc = ERR_TYPE_MISMATCH;
	}
	else {
		/* make sure the larger data type is in 'val1'. */
		int t2_is_larger = (typecmp(t1, t2) < 0);
		if (t2_is_larger) {
			/* swap the values and types. */
			data_type_t t_tmp = t1;
			struct tracker_filter_node* val_tmp = val1;
			val1 = val2;
			val2 = val_tmp;
			t1 = t2;
			t2 = t_tmp;
		}

		result->type = NT_CONST;
		result->data_type = t1;
		RESULT_DATA_TYPE(result) = t1;

		switch (t1) {
			case DT_LONG:
				RESULT_LONG(result) =
					CONST_LONG(val1) & cast_long(val2);
				break;
			case DT_INT:
				RESULT_INT(result) =
					CONST_INT(val1) & cast_int(val2);
				break;
			case DT_SHORT:
				RESULT_SHORT(result) =
					CONST_SHORT(val1) & cast_short(val2);
				break;
			case DT_CHAR:
				RESULT_CHAR(result) =
					CONST_CHAR(val1) & cast_char(val2);
				break;
			default:
				/* shouldn't happen. */
				rc = ERR_TYPE_MISMATCH;
				break;
		}
	}

	return rc;
}

exp_error_t
eval_op_bit_or(struct tracker_filter_node* f_op,
	       struct tracker_filter_node* val1,
	       struct tracker_filter_node* val2,
	       struct tracker_filter_node* result)
{
	exp_error_t rc = ERR_NONE;
	data_type_t t1 = val1->data_type;
	data_type_t t2 = val2->data_type;

	/* we don't support bit operations for some data types. */
	if (t1 == DT_STRING || t1 == DT_BOOL || t1 == DT_FLOAT) {
		rc = ERR_TYPE_MISMATCH;
	}
	if (t2 == DT_STRING || t2 == DT_BOOL || t2 == DT_FLOAT) {
		rc = ERR_TYPE_MISMATCH;
	}
	else {
		/* make sure the larger data type is in 'val1'. */
		int t2_is_larger = (typecmp(t1, t2) < 0);
		if (t2_is_larger) {
			/* swap the values and types. */
			data_type_t t_tmp = t1;
			struct tracker_filter_node* val_tmp = val1;
			val1 = val2;
			val2 = val_tmp;
			t1 = t2;
			t2 = t_tmp;
		}

		result->type = NT_CONST;
		result->data_type = t1;
		RESULT_DATA_TYPE(result) = t1;

		switch (t1) {
			case DT_LONG:
				RESULT_LONG(result) =
					CONST_LONG(val1) | cast_long(val2);
				break;
			case DT_INT:
				RESULT_INT(result) =
					CONST_INT(val1) | cast_int(val2);
				break;
			case DT_SHORT:
				RESULT_SHORT(result) =
					CONST_SHORT(val1) | cast_short(val2);
				break;
			case DT_CHAR:
				RESULT_CHAR(result) =
					CONST_CHAR(val1) | cast_char(val2);
				break;
			default:
				/* shouldn't happen. */
				rc = ERR_TYPE_MISMATCH;
				break;
		}
	}

	return rc;
}

exp_error_t
eval_op_bit_xor(struct tracker_filter_node* f_op,
		struct tracker_filter_node* val1,
		struct tracker_filter_node* val2,
		struct tracker_filter_node* result)
{
	exp_error_t rc = ERR_NONE;
	data_type_t t1 = val1->data_type;
	data_type_t t2 = val2->data_type;

	/* we don't support bit operations for some data types. */
	if (t1 == DT_STRING || t1 == DT_BOOL || t1 == DT_FLOAT) {
		rc = ERR_TYPE_MISMATCH;
	}
	if (t2 == DT_STRING || t2 == DT_BOOL || t2 == DT_FLOAT) {
		rc = ERR_TYPE_MISMATCH;
	}
	else {
		/* make sure the larger data type is in 'val1'. */
		int t2_is_larger = (typecmp(t1, t2) < 0);
		if (t2_is_larger) {
			/* swap the values and types. */
			data_type_t t_tmp = t1;
			struct tracker_filter_node* val_tmp = val1;
			val1 = val2;
			val2 = val_tmp;
			t1 = t2;
			t2 = t_tmp;
		}

		result->type = NT_CONST;
		result->data_type = t1;
		RESULT_DATA_TYPE(result) = t1;

		switch (t1) {
			case DT_LONG:
				RESULT_LONG(result) =
					CONST_LONG(val1) ^ cast_long(val2);
				break;
			case DT_INT:
				RESULT_INT(result) =
					CONST_INT(val1) ^ cast_int(val2);
				break;
			case DT_SHORT:
				RESULT_SHORT(result) =
					CONST_SHORT(val1) ^ cast_short(val2);
				break;
			case DT_CHAR:
				RESULT_CHAR(result) =
					CONST_CHAR(val1) ^ cast_char(val2);
				break;
			default:
				/* shouldn't happen. */
				rc = ERR_TYPE_MISMATCH;
				break;
		}
	}

	return rc;
}

exp_error_t
eval_op_bit_not(struct tracker_filter_node* f_op,
		struct tracker_filter_node* val,
		struct tracker_filter_node* result)
{
	exp_error_t rc = ERR_NONE;
	data_type_t t = val->data_type;

	/* we don't support shift operations for some data types. */
	if (t == DT_STRING || t == DT_BOOL || t == DT_FLOAT) {
		rc = ERR_TYPE_MISMATCH;
	}
	else {
		result->type = NT_CONST;
		result->data_type = val->data_type;
		RESULT_DATA_TYPE(result) = val->data_type;

		switch (result->data_type) {
        		case DT_CHAR:
				RESULT_CHAR(result) = ~CONST_CHAR(val);
				break;
        		case DT_INT:
				RESULT_INT(result) = ~CONST_INT(val);
				break;
        		case DT_SHORT:
				RESULT_SHORT(result) = ~CONST_SHORT(val);
				break;
        		case DT_LONG:
				RESULT_LONG(result) = ~CONST_LONG(val);
				break;
			default:
				/* shouldn't happen. */
				rc = ERR_INVALID_DATA_TYPE;
				break;
		}
	}

	return rc;
}

exp_error_t
eval_op_logic_and(struct tracker_filter_node* f_op,
		  struct tracker_filter_node* val1,
		  struct tracker_filter_node* val2,
		  struct tracker_filter_node* result)
{
	int val1_bool = cast_bool(val1);
	int val2_bool = cast_bool(val2);

	result->type = NT_CONST;
	RESULT_DATA_TYPE(result) = DT_BOOL;
	RESULT_BOOL(result) = (val1_bool && val2_bool ? TF_TRUE : TF_FALSE);

	return ERR_NONE;
}

exp_error_t
eval_op_logic_or(struct tracker_filter_node* f_op,
		 struct tracker_filter_node* val1,
		 struct tracker_filter_node* val2,
		 struct tracker_filter_node* result)
{
	int val1_bool = cast_bool(val1);
	int val2_bool = cast_bool(val2);

	result->type = NT_CONST;
	RESULT_DATA_TYPE(result) = DT_BOOL;
	RESULT_BOOL(result) = (val1_bool || val2_bool ? TF_TRUE : TF_FALSE);

	return ERR_NONE;
}

exp_error_t
eval_logic_not(struct tracker_filter_node* f_op,
	       struct tracker_filter_node* val,
	       struct tracker_filter_node* result)
{
	int val_bool = cast_bool(val);

	result->type = NT_CONST;
	RESULT_DATA_TYPE(result) = DT_BOOL;
	RESULT_BOOL(result) = (val_bool ? TF_FALSE : TF_TRUE);

	return ERR_NONE;
}

exp_error_t
eval_pattern_match(struct tracker_filter_node* f_op,
	           struct tracker_filter_node* val1,
	           struct tracker_filter_node* val2,
	           struct tracker_filter_node* result)
{
	char* str = cast_string(val1);
	char* pat = cast_string(val2);

	result->type = NT_CONST;
	RESULT_DATA_TYPE(result) = DT_BOOL;
	RESULT_BOOL(result) = pattern_match(str, pat);

	if (str)
		mem_free(str);
	if (pat)
		mem_free(pat);

	return ERR_NONE;
}

/* calculate the operator found in the given node */
/* on the two results.				  */
exp_error_t
eval_filter_node_op(struct tracker_filter_node* f_op,
		    struct tracker_filter_node* val1,
		    struct tracker_filter_node* val2,
		    struct tracker_filter_node* result)
{
	exp_error_t rc = ERR_NONE;

	switch(f_op->u.compound.op) {
		case OT_EQUAL:
			rc = eval_op_equal(f_op, val1, val2, result);
			break;
		case OT_LESS:
			rc = eval_op_less(f_op, val1, val2, result);
			break;
		case OT_GREATER:
			rc = eval_op_greater(f_op, val1, val2, result);
			break;
		case OT_LESS_EQUAL:
			rc = eval_op_less_equal(f_op, val1, val2, result);
			break;
		case OT_GREATER_EQUAL:
			rc = eval_op_greater_equal(f_op, val1, val2, result);
			break;
		case OT_NOT_EQUAL:
			rc = eval_op_not_equal(f_op, val1, val2, result);
			break;
		case OT_PLUS:
			rc = eval_op_plus(f_op, val1, val2, result);
			break;
		case OT_MINUS:
			rc = eval_op_minus(f_op, val1, val2, result);
			break;
		case OT_BIT_SHIFT_LEFT:
			rc = eval_op_shift_left(f_op, val1, val2, result);
			break;
		case OT_BIT_SHIFT_RIGHT:
			rc = eval_op_shift_right(f_op, val1, val2, result);
			break;
		case OT_BIT_AND:
			rc = eval_op_bit_and(f_op, val1, val2, result);
			break;
		case OT_BIT_OR:
			rc = eval_op_bit_or(f_op, val1, val2, result);
			break;
		case OT_BIT_XOR:
			rc = eval_op_bit_xor(f_op, val1, val2, result);
			break;
		case OT_BIT_NOT:
			rc = eval_op_bit_not(f_op, val1, result);
			break;
		case OT_LOGIC_AND:
			rc = eval_op_logic_and(f_op, val1, val2, result);
			break;
		case OT_LOGIC_OR:
			rc = eval_op_logic_or(f_op, val1, val2, result);
			break;
		case OT_LOGIC_NOT:
			rc = eval_logic_not(f_op, val1, result);
			break;
		case OT_PATTERN_MATCH:
			rc = eval_pattern_match(f_op, val1, val2, result);
			break;
		default:
			/* shouldn't happen. */
			rc = ERR_INVALID_OPERATOR;
			break;
	}

	return rc;
}

/* evaluate the given compound node recursively. */
/* first recursively calculate each sub-node,    */
/* and then calculate its operator.              */
/* TODO - add short-circuiting support.          */
exp_error_t
eval_filter_node_compound(struct tracker_filter_node* f,
			  struct tracker_filter_node* result)
{
	struct tracker_filter_node result1;
	struct tracker_filter_node result2;
	struct tracker_filter_node* f1;
	struct tracker_filter_node* f2;
	exp_error_t rc1 = ERR_NONE;
	exp_error_t rc2 = ERR_NONE;
	exp_error_t rc = ERR_NONE;

	f1 = f->u.compound.node1;
	rc1 = do_eval_filter_node(f1, &result1);
	if (rc1 == ERR_NONE) {
	        oper_type_t op = f->u.compound.op; 
	        f2 = f->u.compound.node2;
		/* the second sub-node is used only for binary operators. */
		if (!f2) {
		        rc = rc1;
		        goto eval_done;
		}
                /* are we evaluating a short circuiting operator? (logical or or logical and)) */
	        if (op == OT_LOGIC_OR || op == OT_LOGIC_AND) {
		        /* we might be able to short citcuit - 
		           for or, if the lhs is true, we return true for the whole expression
		           for and, if the lhs if false, we return false for the whole expression  */
		        result->type = NT_CONST;
			RESULT_DATA_TYPE(result) = DT_BOOL;
			if (op == OT_LOGIC_OR &&  result1.type == NT_CONST && 
			        cast_bool(&result1) == TF_TRUE){
			        PPRINT(KERN_DEBUG "\nshort circuiting evaluation!\n"); 
				RESULT_BOOL(result) = TF_TRUE;
				rc  = rc1;
				goto eval_done;
			} else if (op == OT_LOGIC_AND &&  result1.type == NT_CONST && 
				   cast_bool(&result1) == TF_FALSE) {
			         PPRINT(KERN_DEBUG "\nshort circuiting evaluation!\n"); 
			         RESULT_BOOL(result) = TF_FALSE;
				 rc  = rc1;
				 goto eval_done;
			}
		}
		/* evaluate the second node */
		rc2 = do_eval_filter_node(f2, &result2);
		if (rc2 == ERR_NONE) {
                         /* calculate the operator on the two results. */
			rc = eval_filter_node_op(f, &result1, &result2, result);
		}
		else {
			rc = rc2;
		}
	}
	else {
		rc = rc1;
	}
eval_done:
	return rc;
}

/* evaluate the given filter node. set the result */
/* in the given result node.			  */
int
do_eval_filter_node(struct tracker_filter_node* f,
		    struct tracker_filter_node* result)
{
	switch (f->type) {
		case NT_VAR:
			return eval_filter_node_var(f, result);
		case NT_CONST:
			return eval_filter_node_const(f, result);
		case NT_COMPOUND:
			return eval_filter_node_compound(f, result);
		default:
			/* shouldn't happen. */
			return 0;
	}
}

/* evaluate the given filter node. */
/* since this is a generic function, it returns a */
/* filter node, containing a constant and the     */
/* proper result.                                 */
exp_error_t
eval_filter_node(filter_node_t* f, filter_node_t* result)
{
	/* assume failure. */
	exp_error_t rc = ERR_INVALID_NODE;

	if (f && result) {
		rc = do_eval_filter_node(f, result);
	}

	return rc;
}
