/*
 *$Id: filter_vars.c,v 1.1 2001/09/14 14:58:43 chooga Exp $
 */

#include "mem_manager_public.h"
#include "filter_vars_public.h"

/* constants. */
#define MAX_PARAM_NUM 5		/* highest supported parameter. */

/* list of variables data. */
static void* var_values[VT_LAST_TYPE];
static int var_initialized = 0;

/*
 * Private functions.
 */

/* get the value of the given parameter */
filter_node_t*
get_param(int param_index)
{
	filter_node_t* val = NULL;

	if (var_initialized && (param_index >= 0 && param_index < MAX_PARAM_NUM)) {
		if (var_values[VT_PARAM]) {
			filter_node_t** params =
					(filter_node_t**)var_values[VT_PARAM];
			val = params[param_index];
		}
	}

	return val;
}

/* set the value of the given parameter to the given value. */
void
set_param(int param_index, filter_node_t* param_val)
{
	/* range checking. */
	if (param_index < 0 || param_index >= MAX_PARAM_NUM)
		return;

	if (!var_values[VT_PARAM]) {
		filter_node_t** params;
		int i;

		var_values[VT_PARAM] =
			mem_alloc(sizeof(filter_node_t*) * MAX_PARAM_NUM);
		params = (filter_node_t**)var_values[VT_PARAM];
		if (!params)
                        return;
		for (i=0; i<MAX_PARAM_NUM; i++)
			params[i] = NULL;
	}
	if (var_values[VT_PARAM]) {
		filter_node_t** params = (filter_node_t**)var_values[VT_PARAM];
		if (params[param_index]) {
			free_filter_node(params[param_index]);
		}
		params[param_index] = param_val;
	}
}

/* free all allocated parameter values. */
void
free_params_list(void)
{
	filter_node_t** params = (filter_node_t**)var_values[VT_PARAM];
	if (params) {
		int i;
		for (i=0; i<MAX_PARAM_NUM; i++) {
			free_filter_node(params[i]);
		}
		mem_free(params);
	}
}

/* initialized the variables array. */
void
init_vars(void)
{
	int i;

	for (i=0; i<VT_LAST_TYPE; i++)
		var_values[i] = NULL;

	var_initialized = 1;
}


/*
 * Kernal variables manipulation code.
 */

/* empty variable values array. */
void
reset_all_vars(void)
{
	int i;

	if (!var_initialized) {
		init_vars();
	}

	for (i=0; i<VT_LAST_TYPE; i++) {
		switch (i) {
			case VT_PARAM:
				free_params_list();
				var_values[i] = NULL;
				break;
			default:
				free_filter_node((filter_node_t*)var_values[i]);
				var_values[i] = NULL;
				break;
		}
	}
}

/* store a given variable's value. */
void set_var(var_type_t type, filter_node_t* val)
{
	if (type > 0 && type < VT_LAST_TYPE && var_initialized) {
		switch (type) {
			case VT_PARAM:
				/* this is an indexed variable. */
				break;
			default:
				var_values[type] = val;
				break;
		}
	}
}

void set_var_indexed(var_type_t type, int var_index, filter_node_t* val)
{
	if (type > 0 && type < VT_LAST_TYPE && var_initialized) {
		switch (type) {
			case VT_PARAM:
				set_param(var_index, val);
				break;
			default:
				/* these are not indexed variables. */
				break;
		}
	}
}

/* evaluate the value of the given parameter, assuming it was previously */
/* set with set_var/set_var_index.                                       */
filter_node_t* get_var(var_type_t type)
{
	filter_node_t* val = NULL;

	if (type > 0 && type < VT_LAST_TYPE && var_initialized) {
		switch(type) {
			case VT_PARAM:
				/* VT_PARAM is an indexed var. */
				break;
			default:
				val = var_values[type];
				break;
		}
	}

	return val;
}

filter_node_t* get_var_indexed(var_type_t type, int var_index)
{
	filter_node_t* val = NULL;

	if (type > 0 && type < VT_LAST_TYPE && var_initialized) {
		switch(type) {
			case VT_PARAM:
				val = get_param(var_index);
				break;
			default:
				/* these are not indexed vars */
				break;
		}
	}

	return val;
}

