/*
 *$Id: main.c,v 1.1 2001/09/14 14:58:43 chooga Exp $
 */
#include <stdio.h>
#include <sys/time.h>
#include <unistd.h>


#include "rule_engine_public.h"
#include "filter_vars_public.h"


/* filter evaluation testing */
void
test_filter_eval_bit(void)
{
	filter_node_t* result = make_bool_val(TF_FALSE);
	exp_error_t rc;

	{
		filter_node_t* f0 = make_long_val(120L);
		filter_node_t* f1 = make_long_val(30L);
		filter_node_t* f2 = make_bit_and(f0, f1);

		printf("\n");
		printf("Evaluating filter:\n");
		print_filter_node(f2);
		rc = eval_filter_node(f2, result);
		printf("eval_filter_node returned '%d'\n", rc);
		if (rc == ERR_NONE) {
			printf("result:\n");
			print_filter_node(result);
		}
		printf("\n");
		free_filter_node(f2);
	}
	{
		filter_node_t* f0 = make_long_val(120L);
		filter_node_t* f1 = make_long_val(30L);
		filter_node_t* f2 = make_bit_or(f0, f1);

		printf("\n");
		printf("Evaluating filter:\n");
		print_filter_node(f2);
		rc = eval_filter_node(f2, result);
		printf("eval_filter_node returned '%d'\n", rc);
		if (rc == ERR_NONE) {
			printf("result:\n");
			print_filter_node(result);
		}
		printf("\n");
		free_filter_node(f2);
	}
	{
		filter_node_t* f0 = make_long_val(120L);
		filter_node_t* f1 = make_long_val(30L);
		filter_node_t* f2 = make_bit_xor(f0, f1);

		printf("\n");
		printf("Evaluating filter:\n");
		print_filter_node(f2);
		rc = eval_filter_node(f2, result);
		printf("eval_filter_node returned '%d'\n", rc);
		if (rc == ERR_NONE) {
			printf("result:\n");
			print_filter_node(result);
		}
		printf("\n");
		free_filter_node(f2);
	}

	free_filter_node(result);
}

int
main(int argc, char* argv[])
{
	struct timeval t1, t2;
	filter_node_t* result = make_bool_val(TF_FALSE);
	exp_error_t rc;
	filter_node_t* f0 = make_string_val("hello world");
	filter_node_t* f1 = make_string_val("lo w");
	filter_node_t* f2 = make_pattern_match(f0, f1);
	filter_node_t* f3 = make_int_val(15);
	filter_node_t* f4 = make_logical_and(f2, f3);
	rule_list_t* rules;

	printf("Printing filter node.\n");
	print_filter_node(f4);


	gettimeofday(&t1, NULL);
	rc = eval_filter_node(f4, result);
	gettimeofday(&t2, NULL);

	printf("   time before eval: %ld, %ld\n", t1.tv_sec, t1.tv_usec);
	printf("   time after  eval: %ld, %ld\n", t2.tv_sec, t2.tv_usec);
	printf("Expression result (rc = %d):\n", rc);
	print_filter_node(result);

	/* filter serialization/deserializatin testing. */
	{
		unsigned char buf[1000];
		int len = serialize_filter_node(f4, buf, 1000);
		printf("serialized 'f4' into '%d' bytes\n", len);

		if (len > 0) {
			filter_node_t* f_deser =
				deserialize_filter_node(buf, len);
			if (f_deser) {
				printf("'f4' deserialized:\n");
				print_filter_node(f_deser);
				free_filter_node(f_deser);
			}
			else {
				printf("Deserialization of 'f4' failed.\n");
			}
		}
	}
	
	/* tracking rules serialization/de-serialization testing. */
	{
		unsigned char buf[1000];
		action_t* a1 = make_action_log(4);
		int len = serialize_tracking_rule(1, 10, a1, f4, buf, 1000);
		printf("serialized rule 1 into '%d' bytes\n", len);

		if (len > 0) {
			action_t* a_deser;
			filter_node_t* f_deser;
			int rule_id;
			int syscall_id;
			int rc = deserialize_tracking_rule(buf,
							   len,
							   &rule_id,
							   &syscall_id,
							   &a_deser,
							   &f_deser);
			if (rc > 0) {
				printf("rule 1 deserialized:\n");
				printf(" rule_id - %d\n", rule_id);
				printf(" syscall_id - %d\n", syscall_id);
				print_action_node(a_deser);
				print_filter_node(f_deser);
				free_filter_node(f_deser);
			}
			else {
				printf("Deserialization of rule 1 failed.\n");
			}
		}
	}

	/* tracking rules add/delete testing. */
	{
		action_t* a1 = make_action_log(4);
		action_t* a2 = make_action_disable_rule(6);
		int rc;
		
		printf("creating tracking rules list\n");
		rules = create_tracking_rule_list();
		printf("adding tracking rules\n");
		rc = add_tracking_rule(rules, 6, f4, a1);
		rc = add_tracking_rule(rules, 3, result, a2);

		print_tracking_rules(rules);

		
		printf("deleting all tracking rules\n");
		del_all_tracking_rules_in_list(rules);
		print_tracking_rules(rules);
		delete_tracking_rule_list(rules);
	}

	/* kernel variables assignment testing. */
	{
		exp_error_t rc;
		filter_node_t* v_sys_id = make_int_val(10);
		filter_node_t* v_param_0 = make_string_val("/etc/passwd");
		filter_node_t* f0 = make_var_val(VT_SYSCALL_ID, DT_INT);
		filter_node_t* f1 =
			make_var_indexed_val(VT_PARAM, 0, DT_STRING);
		filter_node_t* result = make_bool_val(TF_FALSE);

		reset_all_vars();
		set_var(VT_SYSCALL_ID, v_sys_id);
		set_var_indexed(VT_PARAM, 0, v_param_0);

		rc = eval_filter_node(f0, result);
		printf("syscall_id (%d):\n", rc);
		print_filter_node(result);

		rc = eval_filter_node(f1, result);
		printf("param 0 (%d):\n", rc);
		print_filter_node(result);
	}
	/* check short circuiting */
	{
	    filter_node_t* f0 = make_int_val(1);
	    filter_node_t* f1 = make_int_val(0);
	    filter_node_t* f_and = make_logical_and(f0, f1);
	    filter_node_t* f_or = make_logical_or(f0, f1);

	    printf("-------------------\n");
	    rc = eval_filter_node(f_and, result);
	    
	    printf("short circuiting support test: and - '%d'\n",rc);
	    print_filter_node(result);
 	
	    rc = eval_filter_node(f_or, result);
	    printf("short circuiting support test: or - '%d'\n",rc);
	    print_filter_node(result);
	}	

	test_filter_eval_bit();
	
	free_filter_node(result);

	return 0;
}
