/*
 *$Id: rule_engine_private.h,v 1.1 2001/09/14 14:58:43 chooga Exp $
 */
#ifndef _RULE_ENGINE_PRIVATE_H
#define _RULE_ENGINE_PRIVATE_H

/* printing macro. */
#include "archdep.h"

/*
 * Private data types of the rules matching engine.
 */

#include "rule_engine_public.h"

/*
 * data types used by the tracking module.
 */

/* types of supported nodes. */
typedef enum {
	NT_NIL = 0,
	NT_VAR,
	NT_CONST,
	NT_COMPOUND,
} node_type_t;

/* types of supported operators. */
typedef enum {
	OT_EQUAL = 0,
	OT_LESS,
	OT_GREATER,
	OT_LESS_EQUAL,
	OT_GREATER_EQUAL,
	OT_NOT_EQUAL,
	OT_PLUS,
	OT_MINUS,
	OT_BIT_SHIFT_LEFT,
	OT_BIT_SHIFT_RIGHT,
	OT_BIT_AND,
	OT_BIT_OR,
	OT_BIT_XOR,
	OT_BIT_NOT,
	OT_LOGIC_AND,
	OT_LOGIC_OR,
	OT_LOGIC_NOT,
	OT_PATTERN_MATCH,
	OT_LAST_OPERATOR
} oper_type_t;

/* operators symbols, used for printing. */
extern const char* operators_symbols[];

/* 'var' tracker filter node - contains a single variable. */
struct tracker_filter_node_var {
	var_type_t type;	/* type of variable.              */
	int var_index;		/* index, if an indexed variable. */
};

/* 'constant' tracker filter node - contains a single constant. */
struct tracker_filter_node_const {
	data_type_t type;	/* type of constant.              */
	union {
		int b;		/* bool   */
		char c;		/* char   */
		int i;		/* int    */
		long l;		/* int    */
		short s;	/* short  */
		float f;	/* float  */
		char* str;	/* string */
	} u;
};

/* constant filter node access macros. */
#define	CONST_DATA_TYPE(cs)	((cs)->u.constant.type)
#define	CONST_BOOL(cs)		((cs)->u.constant.u.b)
#define	CONST_CHAR(cs)		((cs)->u.constant.u.c)
#define	CONST_SHORT(cs)		((cs)->u.constant.u.s)
#define	CONST_INT(cs)		((cs)->u.constant.u.i)
#define	CONST_LONG(cs)		((cs)->u.constant.u.l)
#define	CONST_FLOAT(cs)		((cs)->u.constant.u.f)
#define	CONST_STRING(cs)	((cs)->u.constant.u.str)

/* 'compound' tracker filter node - contains operator and one or more nodes. */
struct tracker_filter_node_compound {
	oper_type_t op;			   /* type of operator. 	      */
	struct tracker_filter_node* node1; /* first node.     		      */
	struct tracker_filter_node* node2; /* second node - binary operators. */
};

/* tracker filter structure. */
struct tracker_filter_node {
	node_type_t type;		/* type of node. */
	data_type_t data_type;		/* data type of node, as inherited. */
					/* from its contained nodes.	    */
	union {
		struct tracker_filter_node_var var;
		struct tracker_filter_node_const constant;
		struct tracker_filter_node_compound compound;
	} u;
};

/* types of actions that may be taken. */
typedef enum {
	AT_NIL = 0,
	AT_LOG,
	AT_RULE_ENABLE,
	AT_RULE_DISABLE,
	AT_PROCESS_SUSPEND,
	AT_PROCESS_KILL,
	AT_SYSCALL_FAIL,
	AT_LAST_ACTION_TYPE
} action_type_t;

/* tracker action structure. */
struct tracker_action {
	action_type_t type;
	union {
		int priority;	/* priority for message logging.      */
		int rule_id;	/* rule ID to be enabled/disabled.    */
		int error_code; /* error code when failing a syscall. */
	} u;
};

/* tracker rule structure. */
struct tracker_rule {
	int rule_id;
	struct tracker_filter_node* filter;
	struct tracker_action action;
};

/* node building function - used by various node-building routines. */
struct tracker_filter_node*
make_compound_node_with_type(oper_type_t op,
		   	     struct tracker_filter_node* node1,
		   	     struct tracker_filter_node* node2,
		   	     data_type_t data_type);

/* filter node serialize/deserialize forward declarations. */
int do_serialize_filter_node(filter_node_t* f,
		             unsigned char* buf,
		             int buflen,
		             unsigned char** pos);
int do_deserialize_filter_node(filter_node_t** pf,
		               unsigned char* buf,
		               int buflen,
		               unsigned char** pos);

#endif /* _RULE_ENGINE_PRIVATE_H */
