#!/usr/bin/perl

#
# $Id: gen_syscalls.pl,v 1.1 2001/09/14 15:15:01 chooga Exp $
#

# input/output files.
$hijack_stubs_file = "syscall_hijack_autogen.c";
$tracker_stubs_file = "syscall_tracker_autogen.c";
$data_file = "syscalls.dat";

# data types mappings.
%types_map = (
  'char', 'char',
  'uchar', 'unsigned char',
  'schar', 'signed char',
  'short', 'short',
  'int', 'int',
  'long', 'long',
  'string', 'char*',
  'const_string', 'const char*',
  'pointer', 'void*',
  'mode_t', 'mode_t',
  'size_t', 'size_t',
  'pid_t', 'pid_t',
);

%types_makeval_map = (
  'char', 'make_char_val',
  'short', 'make_short_val',
  'int', 'make_int_val',
  'long', 'make_long_val',
  'string', 'make_string_val',
  'const_string', 'make_string_val',
  'mode_t', 'make_long_val',
  'size_t', 'make_long_val',
  'pid_t', 'make_long_val',
);

%types_specifier_map = (
  'char', 'c',
  'short', 'h',
  'int', 'i',
  'long', 'l',
  'string', 's',
  'const_string', 's',
  'mode_t', 'l',
  'size_t', 'l',
  'pid_t', 'l',
);

$os = `uname -s`;
chop($os);

if ($#ARGV != 0) {
	die "Usage - $0 <syscalls number>";
}

$num_syscalls = $ARGV[0];

$cvs_id = "/*\n * \$Id\$\n */\n\n";

# open data file, scan and check input, generate output along the way.
open(DAT, "<$data_file") ||
	die "Cannot open file '$data_file' for reading - $!";
print "generating system calls stubs - $num_syscalls system calls\n";

if (-e $hijack_stubs_file) {
    rename("$hijack_stubs_file", "${hijack_stubs_file}.old") ||
	    die "Cannot rename '$hijack_stubs_file' to '${hijack_stubs_file}.old' - $!";
}
if (-e $tracker_stubs_file) {
    rename("$tracker_stubs_file", "${tracker_stubs_file}.old") ||
	    die "Cannot rename '$tracker_stubs_file' to '${tracker_stubs_file}.old' - $!";
}
open(HIJACK_AUTOGEN, ">$hijack_stubs_file") ||
	die "Cannot open '$hijack_stubs_file' for writing - $!";
open(TRACKER_AUTOGEN, ">$tracker_stubs_file") ||
	die "Cannot open '$tracker_stubs_file' for writing - $!";

print HIJACK_AUTOGEN $cvs_id;
print HIJACK_AUTOGEN "/* This file is automatically generated - do not edit. */\n";
print HIJACK_AUTOGEN "/* built by '$0 $num_syscalls' for $os. */\n\n";

if ($os eq "Linux") {
    print HIJACK_AUTOGEN "\n#include <asm/errno.h>\n\n";
} elsif ($os eq "FreeBSD") {
    print HIJACK_AUTOGEN "\n#include <sys/types.h>\n#include <sys/errno.h>\n#include <sys/sysproto.h>\n\n";
}

print TRACKER_AUTOGEN $cvs_id;
print TRACKER_AUTOGEN "/* This file is automatically generated - do not edit. */\n";
print TRACKER_AUTOGEN "/* built by '$0 $num_syscalls' for $os. */\n\n";
if ($os eq "Linux") {
    print TRACKER_AUTOGEN "\n#include <asm/errno.h>\n\n";
} elsif ($os eq "FreeBSD") {
    print TRACKER_AUTOGEN "\n#include <sys/types.h>\n#include <sys/errno.h>\n#include <sys/sysproto.h>\n\n";
}

@syscall_ids = ();
@syscall_functions = ();

$first_stub_function = 'yes';

while (<DAT>) {
	chop;
	s/#.*$//;               # strip comments.
	/^\s*$/ && next;	# skip empty lines.
	# handle 'include' lines:
	if (/^\s*include\s*:\s*([^\s]+)\s*$/) {
		$include_file = $1;
		print HIJACK_AUTOGEN "#include $include_file\n";
		next;
	}
	# handle 'asis' lines:
	if (/^\s*asis:\s*(.*)$/) {
		$asis = $1;
		print HIJACK_AUTOGEN "$asis", "\n";
		next;
	}
	# handle 'cpp' lines:
	if (/^\s*cpp:\s*(.*)$/) {
		$cpp = $1;
		print HIJACK_AUTOGEN "#${cpp}", "\n";
		next;
	}
	# syscall format:
	#kernel_function: syscall_ID: syscall_name(param_type param_name...);
	if (/^\s*([^:\s]+)\s*:\s*([^:\s]+)\s*:\s*([^\(\s]+)\s*\(([^\)]+)\)\s*;\s*$/) {
		$kernel_function = $1;
		$syscall_id = $2;
		$syscall_name = $3;
		$syscall_params = $4;
		push(@syscall_ids, $syscall_id);
		push(@syscall_functions, $syscall_name);
		print "kernel_function - '$kernel_function'\n";
		print "    syscall_id - '$syscall_id'\n";
		print "    syscall_name - '$syscall_name'\n";
		#print "    syscall_params - '$syscall_params'\n";
		# break down the parameters list.
		@syscall_params_strings = split(',', $syscall_params);
		@syscall_param_matchable = ();
		@syscall_param_types = ();
		@syscall_param_names = ();
		@syscall_param_fmt_spec = ();
		foreach $syscall_param_string (@syscall_params_strings) {
			if ($syscall_param_string =~ /^\s*(matchable|unmatchable)\s+([^\s]+)\s+([^\s]+)\s*$/) {
				$param_matchable = $1;
				$param_type = $2;
				$param_name = $3;
				push(@syscall_param_matchable, $param_matchable);
				push(@syscall_param_types, $param_type);
				push(@syscall_param_names, $param_name);
				# avoid passing unknown types to the logging function by not encoding 
				# them in the param format spec
				if (defined($types_specifier_map{$param_type})){
				        push(@syscall_param_fmt_spec,$types_specifier_map{$param_type});
				}
			}
			else {
				die "Invalid parameters syntax in '$data_file' line $.";
			}
		}
		print "    params num: ", $#syscall_param_types + 1, "\n";
		for ($i = 0; $i <= $#syscall_param_types; $i++) {
			print "     param $i: "
				. ($syscall_param_matchable[$i] eq 'matchable' ? "$matchable" : "")
				. "type '$syscall_param_types[$i]'"
				. " name '$syscall_param_names[$i]'\n";
		}
		# generate syscall stub function.
		if ($first_stub_function eq 'yes') {
			print HIJACK_AUTOGEN "/*\n";
			print HIJACK_AUTOGEN " * Stub functions for syscall hijacking.\n";
			print HIJACK_AUTOGEN " */\n\n";
			print TRACKER_AUTOGEN "/*\n";
			print TRACKER_AUTOGEN " * Tracking functions for syscall hijacking.\n";
			print TRACKER_AUTOGEN " */\n\n";
			$first_stub_function = 'no';
		}
		print HIJACK_AUTOGEN
			"/* stub function for syscall '$syscall_name' ($syscall_id). */\n";
		if ($os eq "Linux") {
			print HIJACK_AUTOGEN "asmlinkage ";
		}
		print HIJACK_AUTOGEN
			"static int\nstub_syscall_${syscall_id}(";
		if ($os eq "Linux") {
		    $param_output_string = '';
		    $param_names_string = '';
		    for ($i = 0; $i <= $#syscall_param_types; $i++) {
		    	    $param_type = $syscall_param_types[$i];
			    $param_name = $syscall_param_names[$i];
			    $param_output_type = $types_map{$param_type};
			    if (!defined($param_output_type) || $param_output_type eq '') {
				    die "Invalid parameter type '$param_type' "
					    . "in '$data_file' line $.";
			    }
			    $param_output_string .= "$param_output_type $param_name";
			    if ($i < $#syscall_param_types) {
			        $param_output_string .=  ", ";
			    }
			    if ($param_type =~ /^string$|^const_string$/){
			            $param_names_string .= "kern_";
			    }
			    if (defined($types_specifier_map{$param_type})) {
                                $param_names_string .= "$param_name";
				if ($i < $#syscall_param_types) {
				    $param_names_string .= ", ";
				}
			    }
		    }
		} elsif ($os eq "FreeBSD") {
		    $param_output_string = "struct proc *proc, struct ${syscall_name}_args *uap";
		}
		print HIJACK_AUTOGEN "${param_output_string}) {\n";
		print HIJACK_AUTOGEN "\ttypedef ";
		print HIJACK_AUTOGEN "int\n\t\t(*stub_func_${syscall_id})($param_output_string);\n";
		print HIJACK_AUTOGEN "\tstub_func_${syscall_id} func;\n";
		print HIJACK_AUTOGEN "\tint retval;\n";
		if ($os eq "Linux") {
		    print HIJACK_AUTOGEN "\tint rc;\n";
		}

		# inc reference count.
		if ($os eq "Linux") {
            	    print HIJACK_AUTOGEN "\trc = down_interruptible(&syscall_sem[$syscall_id]);\n";
            	    print HIJACK_AUTOGEN "\tif (rc < 0)\n";
            	    print HIJACK_AUTOGEN "\t\treturn rc;\n";
		}
                print HIJACK_AUTOGEN "\tsyscall_ref_count[$syscall_id]++;\n";
		if ($os eq "Linux") {
            	    print HIJACK_AUTOGEN "\tup(&syscall_sem[$syscall_id]);\n";
		}
		
		print HIJACK_AUTOGEN "\tfunc = (stub_func_${syscall_id})hijacking_functions[$syscall_id];\n";
		print HIJACK_AUTOGEN "\n";
		if ($os eq "Linux") {
		    print HIJACK_AUTOGEN "\tretval = (func ? "
			    . "(*func)("
			    . join(', ', @syscall_param_names)
			    . ") : -ENOSYS);\n";
		} elsif ($os eq "FreeBSD") {
		    print HIJACK_AUTOGEN "\tretval = (func ? "
		    	    . "(*func)(proc, uap) : -ENOSYS);\n";
		}

		# dec reference count.
		if ($os eq "Linux") {
            	    print HIJACK_AUTOGEN "\trc = down_interruptible(&syscall_sem[$syscall_id]);\n";
            	    print HIJACK_AUTOGEN "\tif (rc < 0)\n";
            	    print HIJACK_AUTOGEN "\t\treturn rc;\n";
		}
                print HIJACK_AUTOGEN "\tsyscall_ref_count[$syscall_id]--;\n";
		if ($os eq "Linux") {
            	    print HIJACK_AUTOGEN "\tup(&syscall_sem[$syscall_id]);\n";
		}
		print HIJACK_AUTOGEN "\n";
		print HIJACK_AUTOGEN "\treturn retval;\n";

		print HIJACK_AUTOGEN "}\n\n";
		
		# generate a track function for the tracker module.
		print TRACKER_AUTOGEN
			"/* tracking function for syscall '$syscall_name' ($syscall_id). */\n";
		if ($os eq "Linux") {
			print TRACKER_AUTOGEN "asmlinkage ";
		}
		print TRACKER_AUTOGEN
			"static int\ntrack_syscall_${syscall_id}(";
		print TRACKER_AUTOGEN "${param_output_string}) {\n";
		print TRACKER_AUTOGEN "\ttypedef ";
		print TRACKER_AUTOGEN "int\n\t\t(*track_func_${syscall_id})($param_output_string);\n";
		print TRACKER_AUTOGEN "\ttrack_func_${syscall_id} func = (track_func_${syscall_id})orig_syscalls[$syscall_id];\n";
		# prepare variables for copying strings from user-space.
		for ($i = 0; $i <= $#syscall_param_types; $i++) {
			$param_type = $syscall_param_types[$i];
			$param_name = $syscall_param_names[$i];
			$param_matchable = $syscall_param_matchable[$i];
			if ($param_matchable eq 'matchable' && $param_type =~ /^string$|^const_string$/) {
				# special handling for string matchables.
				$param_output_type = $types_map{$param_type};
				print TRACKER_AUTOGEN "\tchar* kern_${param_name} = NULL;\n";
			}
		}
		print TRACKER_AUTOGEN "\tlong retval;\n";
		print TRACKER_AUTOGEN "\n";
		# copy strings from user-space.
		for ($i = 0; $i <= $#syscall_param_types; $i++) {
			$param_type = $syscall_param_types[$i];
			$param_name = $syscall_param_names[$i];
			$param_matchable = $syscall_param_matchable[$i];
			if ($param_matchable eq 'matchable' && $param_type =~ /^string$|^const_string$/) {
			    if ($os eq "Linux") {
				print TRACKER_AUTOGEN "\tretval = getstring(${param_name}, &kern_${param_name});\n";
			    } elsif ($os eq "FreeBSD") {
				print TRACKER_AUTOGEN "\tretval = getstring(uap->${param_name}, &kern_${param_name});\n";
			    }
			    print TRACKER_AUTOGEN "\tif (retval < 0) {\n";
			    print TRACKER_AUTOGEN "\t\tkern_${param_name} = NULL;\n";
			    print TRACKER_AUTOGEN "\t\tgoto tracker_${syscall_id}_failed_getstring;\n";
			    print TRACKER_AUTOGEN "\t}\n";
			}
		}

		print TRACKER_AUTOGEN "\n";
		# generate the filtering code.
		print TRACKER_AUTOGEN "\t{\n";
		print TRACKER_AUTOGEN "\t\taction_t* action = NULL;\n";
		# generate function parameters variables.
		for ($i = 0; $i <= $#syscall_param_types; $i++) {
			$param_type = $syscall_param_types[$i];
			$param_name = $syscall_param_names[$i];
			$param_matchable = $syscall_param_matchable[$i];
			$param_makeval = $types_makeval_map{$param_type};
			if ($param_type =~ /^string$|^const_string$/) {
				$param_name = "kern_${param_name}";
			} elsif ($os eq "FreeBSD") {
				$param_name = "uap->${param_name}";
			}
			if ($param_matchable eq 'matchable' && defined($param_makeval) && $param_makeval ne '') {
				print TRACKER_AUTOGEN "\t\tfilter_node_t* param${i} = ${param_makeval}(${param_name});\n";
			}
		}
		# generate process-state variables.
		# uid.
		print TRACKER_AUTOGEN "\t\tfilter_node_t* proc_uid = make_long_val((long)current->uid);\n";
		print TRACKER_AUTOGEN "\t\tfilter_node_t* proc_gid = make_long_val((long)current->gid);\n";
		print TRACKER_AUTOGEN "\t\tfilter_node_t* proc_comm = make_string_val(current->comm);\n";

		print TRACKER_AUTOGEN "\n";
		if ($os eq "Linux") {
		    print TRACKER_AUTOGEN "\t\tretval = down_interruptible(&tracker_sem);\n";
		    print TRACKER_AUTOGEN "\t\tif (retval < 0)\n";
		    print TRACKER_AUTOGEN "\t\t\treturn retval;\n";
		}
		print TRACKER_AUTOGEN "\t\treset_all_vars();\n";
		# generate function parameters variable setting calls.
		for ($i = 0; $i <= $#syscall_param_types; $i++) {
			$param_type = $syscall_param_types[$i];
			$param_name = $syscall_param_names[$i];
			$param_matchable = $syscall_param_matchable[$i];
			$param_makeval = $types_makeval_map{$param_type};
			if ($param_matchable eq 'matchable' && defined($param_makeval) && $param_makeval ne '') {
				print TRACKER_AUTOGEN "\t\tset_var_indexed(VT_PARAM, ${i}, param${i});\n";
			}
		}
		# generate process state variable setting calls.
		print TRACKER_AUTOGEN "\t\tset_var(VT_PROC_UID, proc_uid);\n";
		print TRACKER_AUTOGEN "\t\tset_var(VT_PROC_GID, proc_gid);\n";
		print TRACKER_AUTOGEN "\t\tset_var(VT_PROC_COMM, proc_comm);\n";
		print TRACKER_AUTOGEN "\t\taction = filter_syscall(rules[${syscall_id}], ${syscall_id});\n";
		print TRACKER_AUTOGEN "\t\tif (action) {\n";
		print TRACKER_AUTOGEN "\t\t\ttracker_perform_action(rules[${syscall_id}], action, ${syscall_id}, ";
		print TRACKER_AUTOGEN "\"".join('',@syscall_param_fmt_spec)."\" , ${param_names_string});\n";
		print TRACKER_AUTOGEN "\t\t}\n";
		if ($os eq "Linux") {
		    print TRACKER_AUTOGEN "\t\tup(&tracker_sem);\n";
		}
		print TRACKER_AUTOGEN "\t}\n";


		print TRACKER_AUTOGEN "\n";
		$first_putstring_function = 'yes';
		# release strings that were copied from user-space.
		for ($i = 0; $i <= $#syscall_param_types; $i++) {
			$param_type = $syscall_param_types[$i];
			$param_name = $syscall_param_names[$i];
			$param_matchable = $syscall_param_matchable[$i];
			if ($param_matchable eq 'matchable' && $param_type =~ /^string$|^const_string$/) {
				if ($first_putstring_function eq 'yes') {
				    print TRACKER_AUTOGEN "  tracker_${syscall_id}_failed_getstring:\n";
				    $first_putstring_function = 'no';
				}
				print TRACKER_AUTOGEN "\tif (kern_${param_name})\n";
				print TRACKER_AUTOGEN "\t\tputstring(kern_${param_name});\n";
			}
		}
		if ($first_putstring_function eq 'no') {
		    print TRACKER_AUTOGEN "\n\tif (retval < 0)\n";
		    print TRACKER_AUTOGEN "\t\treturn retval;\n\n";
		}
		
		# generate invocation of the original system call.
		print TRACKER_AUTOGEN "\n";
		if ($os eq "Linux") {
		    print TRACKER_AUTOGEN "\treturn (func ? "
			    . "(*func)("
			    . join(', ', @syscall_param_names)
			    . ") : -ENOSYS);\n";
		} elsif ($os eq "FreeBSD") {
		    print TRACKER_AUTOGEN "\treturn (func ? "
			    . "(*func)(proc, uap) : -ENOSYS);\n";
		}
		print TRACKER_AUTOGEN "}\n\n";
		next;
	}
	else {
		die "Invalid syntax in '$data_file' line $.";
	}
}

print HIJACK_AUTOGEN "\n";
print TRACKER_AUTOGEN "\n";

print HIJACK_AUTOGEN "/* pointers to stub functions - to be used */\n";
print HIJACK_AUTOGEN "/* when hijacking system calls.            */\n";
print HIJACK_AUTOGEN "func_ptr stub_functions[] = {\n";
print TRACKER_AUTOGEN "/* pointers to stub functions - to be used */\n";
print TRACKER_AUTOGEN "/* when hijacking system calls.            */\n";
print TRACKER_AUTOGEN "func_ptr track_functions[] = {\n";
for (0..$num_syscalls) {
	$syscall_id = $_;
	# check if we have a stub function for this syscall.
	for ($i=0; $i <= $#syscall_ids; $i++) {
		if ($syscall_ids[$i] == $syscall_id) {
			last;
		}
	}
	if ($i <= $#syscall_ids) {
		print HIJACK_AUTOGEN "\t(func_ptr)stub_syscall_${_},\n";
		print TRACKER_AUTOGEN "\t(func_ptr)track_syscall_${_},\n";
	}
	else {
		print HIJACK_AUTOGEN "\t(func_ptr)NULL,\n";
		print TRACKER_AUTOGEN "\t(func_ptr)NULL,\n";
	}
}
print HIJACK_AUTOGEN "};\n";
print TRACKER_AUTOGEN "};\n";

close(HIJACK_AUTOGEN) ||
	die "Error closing '$hijack_stubs_file' - $!";
close(TRACKER_AUTOGEN) ||
	die "Error closing '$tracker_stubs_file' - $!";

exit(0);

