/*
 *$Id: module_main.c,v 1.1 2001/09/14 15:15:01 chooga Exp $
 */

#include <linux/module.h>
#include <linux/version.h>
#include <linux/modversions.h>
#include <linux/config.h>

MODULE_AUTHOR("guy keren <choo@actcom.co.il>, mulix <mulix@actcom.co.il>");

MODULE_DESCRIPTION("System Call Filter Evaluation and Userspace Interface, "
		   "for 'syscalltrack'");

#include "filter_vars_public.h"
#include "rule_engine_public.h"
#include "syscall_hijack_public.h"
#include "module_interface.h"

#include <linux/unistd.h>
#include <linux/sys.h>
#include <linux/errno.h>
#include <linux/kernel.h>
#include <linux/malloc.h>
#include <linux/init.h>
#include <linux/sysctl.h>
#include <asm/uaccess.h>
#include <asm/semaphore.h>

#include "sct_compat.h"

/*
 * Forward declarations.
 */
int tracker_sysctl(ctl_table *table, int *name, int namelen,
                   void *oldval, size_t *oldlenp,
                   void *newval, size_t newlen, void **context);

/* array of addresses of original system calls. */
static func_ptr orig_syscalls[NR_syscalls];

/* array of counters of rules registered per syscall.   */
/* used for reference counting, for hijacking purposes. */
static unsigned int src[NR_syscalls];

/* array of rules list, per syscall. */
static rule_list_t* rules[NR_syscalls];

static DECLARE_MUTEX(tracker_sem);

/* data used to register sysctl functions for user-mode access to the module. */
static ctl_table tracker_table[] = {
	{CTL_TRACKER_ADD_RULE,
	 NULL,
	 NULL,
	 1024,
	 0600,
	 NULL,
	 NULL,
	 &tracker_sysctl},
	{CTL_TRACKER_DEL_RULE, NULL, NULL, 1024,
	 0600, NULL, NULL, &tracker_sysctl},
	{CTL_TRACKER_ENABLE_RULE, NULL, NULL, 1024,
	 0600, NULL, NULL, &tracker_sysctl},
	{CTL_TRACKER_DISABLE_RULE, NULL, NULL, 1024,
	 0600, NULL, NULL, &tracker_sysctl},
	{CTL_TRACKER_DEL_ALL_RULES_FOR_SYSCALL, NULL, NULL, 1024,
	 0600, NULL, NULL, &tracker_sysctl},
	{CTL_TRACKER_DEL_ALL_MODULE_RULES, NULL, NULL, 1024,
	 0600, NULL, NULL, &tracker_sysctl},
	{CTL_TRACKER_PRINT_ALL_RULES, NULL, NULL, 1024, 
	 0600, NULL, NULL, &tracker_sysctl}, 
	{0}
};

static ctl_table tracker_top_table[] = {
	{CTL_TRACKER, NULL, NULL, 0, 0500, tracker_table},
	{0}
};


static struct ctl_table_header* tracker_sysctl_header = NULL;

#ifdef MODULE

/* get PAGE_SIZE memory (to copy a string into it, in our case) */
#define __getstring()     (char *)kmalloc(PAGE_SIZE, GFP_KERNEL)
#define putstring(name)   kfree((name))

static inline int do_getstring(const char* str, char* page)
{
	int retval;
	unsigned long len = PAGE_SIZE;

	/* TODO: what the hell does this code do? */
	if ((unsigned long) str >= TASK_SIZE) {
		if (!segment_eq(get_fs(), KERNEL_DS))
			return -EFAULT;
	} else if (TASK_SIZE - (unsigned long) str < PAGE_SIZE)
		len = TASK_SIZE - (unsigned long) str;

	retval = strncpy_from_user((char *)page, str, len);
	if (retval > 0) {
		if (retval < len)
			return 0;
		return -ENAMETOOLONG;
	}

	return retval;
}

int getstring(const char* str, char** kern_str)
{
	int result;
	char *tmp;

	result = -ENOMEM;
	tmp = __getstring();
	if (tmp)  {
		int retval = do_getstring(str, tmp);
		result = retval;
                *kern_str = tmp;
                if (retval < 0) {
                        putstring(tmp);
                        result = retval;
                }
        }

        return result;
}


#include "syscall_tracker_autogen.c"


/* adds a new rule for a given system call. */
int
tracker_sysctl_add_rule(int rule_id,
			int syscall_id,
			filter_node_t* f,
			action_t* a)
{
    int rc;
    int retval = 0;
    int release_sem = 0;
    func_ptr old_func;
    
    printk(KERN_NOTICE "adding tracking rule %d for syscall %d\n",
	   rule_id, syscall_id);
    /* lock to avoid races with other invocations. */
    if (down_interruptible(&tracker_sem)) {
        retval = -EINTR;
        goto err_add_rule_lock;
    }
    release_sem = 1;
    rc = add_tracking_rule(rules[syscall_id], rule_id, f, a);
    if (rc < 0) {
	printk (KERN_NOTICE "add tracking_rule returned '%d'\n", rc);
        retval = -ENOMEM; /* todo: something better? */
        goto err_add_rule_add_rule;
    }
    /* make sure we don't release 'f' and 'a', since they were consumed */
    /* by the 'add_tracking_rule' function.                             */
    f = NULL;
    a = NULL;
    print_tracking_rules(rules[syscall_id]);

    /* increase reference count. hijack if first rule for syscall. */
    if (src[syscall_id]++ == 0) {
	printk(KERN_NOTICE "first rule for syscall - hijacking syscall\n");
        rc = hijack_syscall(syscall_id,
			    (func_ptr)track_functions[syscall_id],
			    &old_func);
        if (rc < 0) {
            printk(KERN_NOTICE "hijack_syscall failed, error '%d'\n",
		   rc); 
            retval = (rc != -ERESTARTSYS ? rc : -EINTR);
	    /* TODO - reduce ref. count? remove rule? */
            goto err_add_rule_hijack;
        }
	
        orig_syscalls[syscall_id] = old_func;
        MOD_INC_USE_COUNT;
    }
    /* all went well. */
    goto ret_ok_add_rule;

  err_add_rule_hijack:
  err_add_rule_add_rule:
  err_add_rule_lock:
  ret_ok_add_rule:
    if (release_sem)
	up(&tracker_sem);
    if (f)
	free_filter_node(f);
    if (a)
	kfree(a);

    return retval;
}

/* deletes an existing rule for a given system call. */
int
tracker_sysctl_del_rule(int rule_id,
			int syscall_id)
{
    int rc;
    int retval = 0;
    int release_sem = 0;

    printk(KERN_NOTICE "tracker_sysctl: deleting tracking rule %d for syscall %d\n",
           rule_id, syscall_id);

    /* lock to avoid races with other invocations. */
    if (down_interruptible(&tracker_sem)) {
        retval = -ERESTARTSYS;
        goto err_del_rule_lock;
    }
    release_sem = 1;

    if (!has_tracking_rule(rules[syscall_id], rule_id)) {
	printk(KERN_NOTICE "tracker_sysctl: rule '%d' not found.\n", rule_id);
	goto err_del_rule_no_such_rule;
    }

    del_tracking_rule(rules[syscall_id], rule_id);
    print_tracking_rules(rules[syscall_id]);

    if (--src[syscall_id] == 0) {
	printk(KERN_NOTICE "last rule for syscall - releasing syscall\n");
        rc = release_syscall(syscall_id);
        if (rc < 0) {
            printk(KERN_NOTICE "release_syscall failed, error '%d'\n",
		   rc); 
            retval = (rc != -ERESTARTSYS ? rc : -EINTR);
            goto err_del_rule_release;
        }

        orig_syscalls[syscall_id] = (func_ptr)NULL;
        MOD_DEC_USE_COUNT;
    }
    /* all went well. */
    goto ret_ok_del_rule;

  err_del_rule_no_such_rule:
  err_del_rule_release:
  err_del_rule_lock:
  ret_ok_del_rule:
    if (release_sem)
	up(&tracker_sem);

    return retval;
}

/* deletes all existing rules for a given system call. */
int
tracker_sysctl_del_all_rules_for_syscall(int syscall_id)
{
    int rc;
    int retval = 0;
    int release_sem = 0;

    printk(KERN_NOTICE "tracker_sysctl: deleting all tracking rules for syscall %d\n",
           syscall_id);

    /* lock to avoid races with other invocations. */
    if (down_interruptible(&tracker_sem)) {
        retval = -ERESTARTSYS;
        goto err_del_all_rules_for_syscall_lock;
    }
    release_sem = 1;

    if (rules[syscall_id]){
	del_all_tracking_rules_in_list(rules[syscall_id]);
	print_tracking_rules(rules[syscall_id]);
    }
    printk(KERN_NOTICE "tracker_sysctl: del all rules for syscall - releasing syscall\n");
    rc = release_syscall(syscall_id);
    if (rc < 0) {
        printk(KERN_NOTICE "tracker_sysctl: release_syscall failed, error '%d'\n", rc); 
        retval = (rc != -ERESTARTSYS ? rc : -EINTR);
        goto err_del_all_rules_for_syscall_release;
    }
    orig_syscalls[syscall_id] = (func_ptr)NULL;
    src[syscall_id] = 0;

    MOD_DEC_USE_COUNT;

    goto ret_ok_del_all_rules_for_syscall_lock;

  err_del_all_rules_for_syscall_release:
  err_del_all_rules_for_syscall_lock:
  ret_ok_del_all_rules_for_syscall_lock:
    if (release_sem)
	up(&tracker_sem);

    return retval;
}

/* deletes all existing rules in the module */
int
tracker_sysctl_del_all_module_rules(void)
{
    int rc;
    unsigned int i;
    int retval = 0;
    int release_sem = 0;

    printk(KERN_NOTICE "tracker_sysctl: deleting all module tracking rules\n");

    /* lock to avoid races with other invocations. */
    if (down_interruptible(&tracker_sem)) {
        retval = -ERESTARTSYS;
        goto err_del_all_module_rules_lock;
    }
    release_sem = 1;

    /* iterate over all syscalls and release the hijacked ones */
    for (i = 0; i < NR_syscalls; ++i){
	/* not hijacked? go on */
	if (!orig_syscalls[i]) 
	    continue;
	
	if (rules[i])
	    del_all_tracking_rules_in_list(rules[i]);
	
	rc = release_syscall(i);
	if (rc < 0) {
	    printk(KERN_NOTICE "tracker_sysctl: release_syscall for syscall '%d' failed, "
		   "error '%d'\n", i, rc); 
	    retval = (rc != -ERESTARTSYS ? rc : -EINTR);
	    goto err_del_all_module_rules_release;
	}
	orig_syscalls[i] = (func_ptr)NULL;
	src[i] = 0;
	MOD_DEC_USE_COUNT;
    }

    goto ret_ok_del_all_module_rules;

  err_del_all_module_rules_lock:
  err_del_all_module_rules_release:
  ret_ok_del_all_module_rules:
    if (release_sem)
	up(&tracker_sem);

   return retval;
}

/* prints all existing rules in the module */
int
tracker_sysctl_print_all_rules(void)
{
    unsigned int i;
    int retval = 0;
    int printed = 0;
    
    /* lock to avoid races with other invocations. */
    if (down_interruptible(&tracker_sem)) {
        retval = -ERESTARTSYS;
        goto err_print_module_rules_lock;
    }
    
    printk (KERN_NOTICE "**************************************\n");
    printk (KERN_NOTICE "syscalltrack - printing defined rules\n");
    for (i = 0; i < NR_syscalls; ++i)
    {
	if (src[i] && rules[i]){
	    printk(KERN_NOTICE "rules for syscall '%d':\n", i);
	    print_tracking_rules(rules[i]);
	    printed = 1;
	}
    }
    if (!printed){
	printk (KERN_NOTICE "no tracking rules registered currently\n");
    }
    printk (KERN_NOTICE "**************************************\n");
    
    up(&tracker_sem);

 err_print_module_rules_lock:
    return retval;
}

/* sysctl handling routine. allows adding/deleting/enabling/disabling rules. */
int tracker_sysctl(ctl_table *table,
		   int *name, int namelen,
                   void *oldval, size_t *oldlenp,
                   void *newval, size_t newlen, void **context)
{
    int i;
    int rule_id;
    int syscall_id;
    action_t* a;
    filter_node_t* f;
    int rc, retval = 0;
    unsigned char* buf = NULL;

    printk(KERN_DEBUG "tracker_sysctl: namelen = '%d', newlen = '%d'\n", namelen, newlen);
    for (i=0; i<namelen; i++) {
	printk(KERN_DEBUG "tracker_sysctl: name[%d] = %d\n", i, name[i]);
    }
    if (newlen <= 0 || namelen != 1) {
	retval = -EINVAL;
	goto err_out;
    }

    buf = kmalloc(newlen, GFP_KERNEL);
    if (!buf){
	retval = -ENOMEM;
	printk(KERN_NOTICE "tracker_sysctl: buffer allocation failed\n");
	goto err_oom;
    }
    if (copy_from_user(buf, newval, newlen)) {
	retval = -EFAULT; /* TODO: enomem? eaccess? */
	printk(KERN_NOTICE "tracker_sysctl: failed to copy from user\n");
	goto err_copy_from_user;
    }
    printk(KERN_NOTICE "tracker_sysctl: copied data from user\n");
    rc = deserialize_tracking_rule(buf, newlen, &rule_id, &syscall_id, &a, &f);
    if (rc <= 0) {
	printk(KERN_NOTICE "tracker_sysctl: failed to deserialize tracking rule, "
	       "err '%d'\n", rc);
	retval = -EINVAL; /* todo: anything better? */
	goto err_deserialize;
    }

    /* sanity checks */
    if (syscall_id < 0 || syscall_id > NR_syscalls) {
        printk(KERN_NOTICE "syscall id %d out of range\n",
	       syscall_id);
	goto err_invalid_syscall_id;
    }

    /* which operation did the user request? */
    switch (name[0]) {
        case(CTL_TRACKER_ADD_RULE):
	    retval = tracker_sysctl_add_rule(rule_id, syscall_id, f, a);
	    break;
	
        case(CTL_TRACKER_DEL_RULE):
	    retval = tracker_sysctl_del_rule(rule_id, syscall_id);
	    break;
	
	case(CTL_TRACKER_DEL_ALL_RULES_FOR_SYSCALL):
	    retval = tracker_sysctl_del_all_rules_for_syscall(syscall_id);
	    break;

        case(CTL_TRACKER_DEL_ALL_MODULE_RULES):
	    retval = tracker_sysctl_del_all_module_rules();
 	    break;

        case(CTL_TRACKER_PRINT_ALL_RULES):
	    retval = tracker_sysctl_print_all_rules();
	    break;

        default:
	    printk(KERN_NOTICE "tracker_sysctl: invalid or unsupported sysctl "
		   "operation %d\n", name[0]);
	    retval = -EINVAL;
    }
    /* all is well - we can return peacefully. */
    goto ret_out;

 err_invalid_syscall_id:
 err_deserialize:
 err_copy_from_user:
 err_oom:
 err_out:
 ret_out:
    if (buf)
        kfree(buf);

    return retval;
}


int
init_module (void)
{
	int i;

	for (i=0; i<NR_syscalls; ++i) {
		orig_syscalls[i] = (func_ptr)NULL;
		src[i] = 0;
		rules[i] = create_tracking_rule_list();
	}
		
	/* register sysctl handling function. */
 	tracker_sysctl_header = register_sysctl_table(tracker_top_table, 0);
	printk(KERN_DEBUG "tracker_sysctl_header - %p\n", tracker_sysctl_header); 

	return 0;
}

void
cleanup_module (void)
{
	int i;

        printk(KERN_NOTICE "sct_rules.o: cleaning up syscall filter module\n");

	/* register sysctl handling function. */
	if (tracker_sysctl_header)
		unregister_sysctl_table(tracker_sysctl_header);

	for (i=0; i<NR_syscalls; ++i) {
		delete_tracking_rule_list(rules[i]);
	}
}

#endif /* MODULE */
